import threading
import queue
import logging

import django.conf

import mini_buildd.config
import mini_buildd.misc
import mini_buildd.call
import mini_buildd.changes
import mini_buildd.gnupg
import mini_buildd.api
import mini_buildd.threads
import mini_buildd.ftpd
import mini_buildd.package
import mini_buildd.packager
import mini_buildd.builder
import mini_buildd.events
import mini_buildd.cron

import mini_buildd.models.daemon
import mini_buildd.models.repository
import mini_buildd.models.chroot
import mini_buildd.models.gnupg
import mini_buildd.models.subscription

LOG = logging.getLogger(__name__)


class Keyrings():
    """Hold/manage all gnupg keyrings (for remotes and all repository uploaders)."""

    def __init__(self, model):
        self._model = model
        self._our_pub_key = self._model.mbd_get_pub_key()
        self._remotes = self._gen_remotes()
        self._uploaders = self._gen_uploaders()
        self._needs_update = False

    def set_needs_update(self):
        self._needs_update = True

    def close(self):
        self._remotes.close()
        for u in list(self._uploaders.values()):
            u.close()

    def _update(self):
        if self._needs_update:
            self.close()
            self.__init__(self._model)

    def get_remotes(self):
        self._update()
        return self._remotes

    def get_uploaders(self):
        self._update()
        return self._uploaders

    def _gen_remotes(self):
        remotes = mini_buildd.gnupg.TmpGnuPG(tmpdir_options={"prefix": "gnupg-remotes-keyring-"})
        # keyring["remotes"]: Remotes keyring to authorize buildrequests and buildresults
        # Always add our own key
        if self._our_pub_key:
            remotes.add_pub_key(self._our_pub_key)
        for r in mini_buildd.models.gnupg.Remote.mbd_get_active_or_auto_reactivate():
            remotes.add_pub_key(r.key)
            LOG.info(f"Remote key added for '{r}': {r.key_long_id}: {r.key_name}")
        return remotes

    def _gen_uploaders(self):
        """All uploader keyrings for each repository."""
        uploaders = {}
        for r in mini_buildd.models.repository.Repository.mbd_get_active():
            uploaders[r.identity] = r.mbd_get_uploader_keyring()
            # Always add our key too for internal builds
            if self._our_pub_key:
                uploaders[r.identity].add_pub_key(self._our_pub_key)
        return uploaders


START_STOP_LOCK = threading.Lock()


class Daemon(mini_buildd.threads.EventThread, metaclass=mini_buildd.misc.Singleton):
    def __init__(self):
        self.model = mini_buildd.models.daemon.get()

        # Vars that are (re)generated when the daemon model is updated
        self.incoming_queue = queue.Queue()

        # p-d-registration uses DEFAULT_FROM_EMAIL django setting for sending emails; there is no other way to set this.
        # As this setting lives in the django "Daemon" instance, there is no way to set it before django is up.
        # This code updates a django setting at runtime: https://docs.djangoproject.com/en/1.7/topics/settings/#altering-settings-at-runtime
        # Some however actually may be changed: https://code.djangoproject.com/ticket/14628
        # We are confident that DEFAULT_FROM_EMAIL belongs to that group; also, this method is always called thread-locked
        django.conf.settings.DEFAULT_FROM_EMAIL = self.model.email_address

        self.keyrings = Keyrings(self.model)

        self.builder = mini_buildd.builder.Builder(self.model.build_queue_size)   # TODO: adapt max parallel builds on_model_change ?
        self.packager = mini_buildd.packager.Packager()
        self.ftpd = mini_buildd.ftpd.FtpD(endpoint=self.model.mbd_get_ftp_endpoint(), queue=self.incoming_queue)
        self.crontab = mini_buildd.cron.Tab()

        super().__init__(subthreads=[self.ftpd, self.packager, self.builder, self.crontab], events=self.incoming_queue, name="daemon")

        # Events
        self.events_queue = mini_buildd.events.Queue(mini_buildd.events.load(), maxlen=3 * (self.model.show_last_packages + self.model.show_last_builds))

        #: Dict of functions to acquire possible attention strings (see ``config.ATTENTION_MENU``).
        self.attention = {
            "events": "",   # No use case (yet)
            "builds": "",   # No use case (yet)
            "repositories": mini_buildd.mdls().repository.Repository.mbd_attention,
            "chroots": mini_buildd.mdls().chroot.Chroot.mbd_attention,
            "builders": mini_buildd.mdls().gnupg.Remote.mbd_attention,
            "crontab": self.crontab.attention,
        }

    def __str__(self):
        return f"{'UP' if self.is_alive() else 'DOWN'}: {self.model}"

    def handshake_message(self):
        return self.model.mbd_gnupg().gpgme_sign(mini_buildd.config.HANDSHAKE_MESSAGE)

    def sync(self):
        self.model.refresh_from_db()
        self.ftpd.endpoint = self.model.mbd_get_ftp_endpoint()

    def join(self, timeout=None):
        super().join()
        self.events_queue.shutdown()
        self.sync()

        # keyrings.close() is not called implicitly; this leaves tmp files around.
        # There should be a nicer way, really...
        try:
            self.keyrings.close()
        except BaseException:
            pass

    def run_event(self, event):
        if mini_buildd.changes.Buildrequest.match(event):
            self.builder.queue.put(event)
        else:
            self.packager.queue.put(event)  # User upload or build result: packager

    def check(self):
        """Global check -- check Daemon instance and implicitly all it's dependencies."""
        mini_buildd.models.daemon.Daemon.Admin.mbd_action(None, (self.model,), "check")

    def get_title(self):
        """Human-readable short title for this Daemon instance."""
        return f"{self.model.identity}@{self.model.hostname}"

    def mbd_start(self):
        with START_STOP_LOCK:
            if not self.is_alive() and self.model.mbd_is_active():
                self.start()

    def mbd_stop(self):
        with START_STOP_LOCK:
            if self.is_alive():
                self.shutdown()
                self.join()
                Daemon.destroy()
